###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
.. versionadded:: v11.0.53

This agent consolidates data from the bookkeeping and the DFC, generates json out of it,
uploads them to a storage.
It also sends the accounting data to OpenSearch, and fill in 2 tables in the StorageUsageDB

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN StorageUsageDumpAgent
  :end-before: ##END StorageUsageDumpAgent
  :dedent: 2
  :caption: StorageUsageDumpAgent options

"""
# # imports
import asyncio
import json
import re
from collections import defaultdict
from datetime import datetime, timezone
from functools import lru_cache, partial
from pathlib import Path
from textwrap import dedent

from opensearchpy import OpenSearch, helpers

from sqlalchemy import create_engine, Column, Integer, String, BigInteger, Boolean, text, delete, Engine
from sqlalchemy.orm import declarative_base

import numpy as np
import pandas as pd
import zstandard
import time
import re
import boto3
from botocore.exceptions import ClientError

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters
from DIRAC.DataManagementSystem.DB.FileCatalogDB import FileCatalogDB


from LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB import OracleBookkeepingDB


def nested_defaultdict():
    return defaultdict(nested_defaultdict)


AGENT_NAME = "DataManagement/StorageUsageDumpAgent"

S3_BUCKET_NAME = "lhcbdirac"
S3_ENDPOINT_URL = "https://s3.cern.ch:443"
OS_ENDPOINT_URL = "os-lhcb-dirac.cern.ch:443/os"


# This are directory which should not be considered
BAD_PREFIXES = {
    "/lhcb/data/2008/",
    "/lhcb/testCfg/",
    "/lhcb/EXPFEB2022/",
    "/lhcb/calib/",
    "/lhcb/certification/",
    "/lhcb/swtest/",
    "/lhcb/grid/user/",
    "/lhcb/grid/wg/",
    "/lhcb/prdata/",
}

STREAM_TO_EVENTTYPE = {
    "ION": 90700000,
    "IONRAW": 90800000,
    "TURBORAW": 90400000,
    "BEAMGAS": 97000000,
    "FULL": 90000000,
    "TURBO": 94000000,
    "HLT2CALIB": 95200000,
    "TURCAL": 95100000,
    "NOBIAS": 96000000,
    "PASSTHROUGH": 98100000,
    "LUMI": 93000000,
    "SMOGPHY": 90300000,
    "LOWMULT": 90600000,
    "EXPRESS": 91000000,
    "CALIB": 95000000,
    "ERRORS": 92000000,
    "HLTONE": 98000000,
    "L0YES": 99000000,
}

TAPE_STORAGE_TYPES = ("ARCHIVE", "RAW", "RDST")

# Define the SQLAlchemy model
Base = declarative_base()


class StorageSummary(Base):
    __tablename__ = "dump_storage_summary"

    SEName = Column(String(255), primary_key=True)
    SESize = Column(BigInteger)
    SEFiles = Column(BigInteger)
    SESite = Column(String(255))
    SEType = Column(String(255))
    IsDisk = Column(Boolean)


# Some folders are "duplicated", with case sensitivity issue
# PHYSICS / physics
# ALICE / alice
# +----------------------------------------+---+
# | Name                                   | t |
# +----------------------------------------+---+
# | /lhcb/data/2008/RAW/LHCb/PHYSICS/21188 | 2 |
# | /lhcb/data/2008/RAW/LHCb/PHYSICS/21192 | 2 |
# | /lhcb/data/2008/RAW/RICH2/ALICE/21933  | 2 |
# | /lhcb/data/2008/RAW/RICH2/ALICE/21936  | 2 |
# | /lhcb/data/2008/RAW/RICH2/ALICE/21937  | 2 |
# | /lhcb/data/2008/RAW/RICH2/ALICE/21938  | 2 |
# +----------------------------------------+---+
# that's why we need the specific mysql collate


class DirectoryMetadata(Base):
    __tablename__ = "dump_directory_metadata"

    Name = Column(String(255), primary_key=True)
    production = Column(Integer)
    filetype = Column(String(255))
    online_stream = Column(String(255))
    maybe_eventtype = Column(Integer)
    ConfigName = Column(String(255))
    ConfigVersion = Column(String(255))
    Description = Column(String(255))
    ProcPath = Column(String(255))
    EventTypeID = Column(String(255))

    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8mb4", "mysql_collate": "utf8mb4_bin"}


async def returnValueOrRaiseAsync(func, *args, **kwargs):
    loop = asyncio.get_running_loop()
    return returnValueOrRaise(await loop.run_in_executor(None, func, *args, **kwargs))


async def get_prod_data_info():
    """
    Dump the DirectoryUsage table from the DFC for production
    data only.
    We get:
    * directory name
    * storage Element ('FakeSE' for logicial size)
    * size in bytes
    * number of files
    """
    columns = ["Name", "SEName", "SESize", "SEFiles"]
    query = """
    select Name, SEName, du.SESize, du.SEFiles
    from FC_DirectoryList d
    USE INDEX (Name)
    JOIN FC_DirectoryUsage du on d.DirID = du.DirID
    JOIN FC_StorageElements s on du.SEID = s.SEID
    where (Name not like '/lhcb/user/%');
    """
    query = dedent(query.strip())
    return columns, await returnValueOrRaiseAsync(FileCatalogDB()._query, query)


async def get_user_data_info():
    # This takes ~30 minutes to run so we don't run it for now
    columns = ["SEName", "SESize", "SEFiles"]
    query = """
    select SEName, sum(du.SESize), sum(du.SEFiles)
    from FC_DirectoryList d
    JOIN FC_DirectoryUsage du on d.DirID = du.DirID
    JOIN FC_StorageElements s on du.SEID = s.SEID
    where (Name like '/lhcb/user/%') group by SEName;
    """
    query = dedent(query.strip())
    return columns, await returnValueOrRaiseAsync(FileCatalogDB()._query, query)


async def get_bk_metadata() -> tuple[list, list]:
    """
    Dump the production informations from the Bookkeeping:
    * ConfigName: MC, LHCb ...
    * ConfigVersion:2016, Collision18, ...
    * Description: Condition description (Mag polarity, Velo position)
    * ProcPath: Real Data/Reco18 ...
    * EventTypeID: 90700000, 90800000, ...
    * filetype: ALLSTREAM.DST, RAW...
    * production: id of the production, (negative if run number)

    Something like:
    * ('MC', '2016', 'Beam6500GeV-2016-MagUp-Nu1.6-25ns-Pythia8',
    'Sim09b/Trig0x6138160F/Reco16/Turbo03/Stripping26NoPrescalingFlagged',
    15266005, 'ALLSTREAMS.MDST', 63445)
    * ('LHCb', 'Collision18', 'Beam6500GeV-VeloClosed-MagDown',
    'Real Data/Reco18/Stripping34/AnaProd-v1r2053-collision_B02DstarpDmKS',
    90000000, 'B02DSTARPDMKS.ROOT', 250478)
    * ('LHCb', 'Collision24', 'Beam2680GeV-VeloClosed-MagUp',
    'Real Data', 95200000, 'RAW', -309550)
    """
    columns = ["ConfigName", "ConfigVersion", "Description", "ProcPath", "EventTypeID", "filetype", "production"]
    query = """
    with procpaths (procpath, id) as (
        SELECT distinct SUBSTR(SYS_CONNECT_BY_PATH(name, '/'), 2), id
        FROM processing
        START WITH id in (select distinct id from processing where parentid is NULL)
        CONNECT BY NOCYCLE PRIOR id=parentid
    ), condesc (daqid, simid, description) as (
        SELECT distinct DAQPERIODID, NULL, DESCRIPTION FROM data_taking_conditions
        UNION
        SELECT distinct NULL, simid, simdescription FROM simulationConditions
    )
    select
        c.configname, c.configversion, condesc.description, procpaths.procpath,
        eventtypes.eventtypeid, ftypes.name as filetype, cont.production
    from productionscontainer cont
    inner join configurations c on cont.configurationid=c.configurationid
    inner join condesc on condesc.daqid=cont.DAQPERIODID or condesc.simid=cont.simid
    inner join procpaths on cont.processingid=procpaths.id
    inner join productionoutputfiles prodoutfiles on prodoutfiles.production=cont.production
    inner join filetypes ftypes on prodoutfiles.filetypeId=ftypes.filetypeid
    inner join eventTypes on eventTypes.EventTypeId=prodoutfiles.eventtypeid
    group by c.configname, c.configversion, condesc.description, procpaths.procpath,
        eventtypes.eventtypeid, ftypes.name, cont.production
    """
    query = dedent(query.strip())
    return columns, await returnValueOrRaiseAsync(OracleBookkeepingDB().dbR_.query, query)


@lru_cache(maxsize=4096)
def se_name_to_type(name: str) -> tuple[str, str]:
    """
    Parses the storage name to return the site and type
    """
    if name.endswith("-EOS"):
        name = name[: -len("-EOS")]
    for t in ["BUFFER", "RAW", "MC-DST", "RDST", "ARCHIVE", "DST", "ANAPROD", "DAQ-EXPORT", "FAILOVER"]:
        if name.endswith(t):
            se_type = t
            se_site = name[: -len(se_type)].strip("-_")
            break
    else:
        se_type = "OTHER"
        # We have a bunch of weird SE's at CERN
        se_site = "CERN" if name.startswith("CERN") else name
    return (se_site, se_type)


@lru_cache(maxsize=4096)
def normalise_eventtype(x: int) -> int:
    """
    Normalizes the event type to group the calibration production event type, eg 90000001 to 90000000
    """
    if not np.isnan(x) and str(int(x)).endswith("1"):
        x = int(str(int(x))[:-1] + "0")
    return x


def find_matching_event_type_row(group):
    """Reduce a group with multiple event types to only the
    one corrosponding to this directory LFN.
    """
    se_name, name = group.name
    # If there is only one entry then there is no mapping to fix
    # e.g. MC where there is only one event type per production
    # e.g. Sprucing where each online stream becomes a different production
    if len(group) == 1:
        return group.iloc[0]
    if len(group["EventTypeID"]) != group["EventTypeID"].nunique():
        raise NotImplementedError(name)
    matches = group["maybe_eventtype"] == group["EventTypeID"]
    if sum(matches) == 0:
        if not group["EventTypeID"].astype(str).str.startswith("9").all():
            raise NotImplementedError("This correction only makes sense for real data")
        # In Run 1 and Run 2 sending to castor instead of offline resulted in an event type ending in a 1
        # See if we get matches after correcting for this
        matches = group["maybe_eventtype"].apply(normalise_eventtype) == group["EventTypeID"].apply(normalise_eventtype)
    if sum(matches) != 1:
        # These corrospond to a tiny fraction of data which is weirdly inconsistent for historical reasons
        # Let them just pick the first match in the group
        if re.search("(?:/FEST/|/2012/RAW/(?:BEAMGAS|HLT1|CALIB)/|/2017/RAW/TEST_MARKUS_CALIB/)", name):
            return group.iloc[0]
        # # TODO: These files are in the DFC but not in the BKK
        if re.search(
            (
                "/(?:133379|133380|133386|133382|133381|133383|"
                "133385|133384|133387|133388|133389|133392|133390|"
                "133391|133394|133395|133393|133398|133399|133400|133396|133397)$"
            ),
            name,
        ):
            return group.iloc[0]
        # If it isn't a known issue, it's better to crash than do the wrong thing
        raise NotImplementedError(name, len(group))
    return group[matches].iloc[0]


def summarise_space(df, group_cols):
    result = nested_defaultdict()
    for _, row in df.groupby(group_cols, dropna=False)["SESize"].sum().reset_index().iterrows():
        result_row = result
        for k in group_cols[:-1]:
            if pd.isna(row[k]) or row[k] == "Unknown":
                if not result["Unknown"]:
                    result["Unknown"] = 0
                result["Unknown"] += int(row["SESize"])
                break
            if k == "ProcPath":
                for k2 in row[k].split("/"):
                    result_row = result_row[k2]
            else:
                result_row = result_row[row[k]]
        else:
            key = row[group_cols[-1]]
            if key in result_row:
                raise NotImplementedError(row)
            result_row[key] = row["SESize"]
    return result2tree(result)


def result2tree(x):
    if isinstance(x, dict):
        return {"children": [{"name": k} | result2tree(v) for k, v in x.items()]}
    elif isinstance(x, int):
        return {"size": x}
    else:
        raise NotImplementedError(x)


def uploadToS3(endpoint_url, bucket_name, access_key, secret_key):
    """
    Uploads to S3 the json data as well as the index.html
    """
    log = gLogger.getSubLogger("uploadToS3")
    log.info("Uploading to S3")
    start_time = time.time()
    s3_client = boto3.client(
        "s3", endpoint_url=endpoint_url, aws_access_key_id=access_key, aws_secret_access_key=secret_key
    )

    try:
        s3_client.upload_file(
            "/tmp/storage.json.zst",
            bucket_name,
            "storage-usage/storage.json.zst",
            ExtraArgs={"ACL": "public-read", "ContentType": "application/zstd"},
        )

    except ClientError as e:
        log.exception("Failed to upload storage.json.zst ", lException=e)

    try:
        s3_client.upload_file(
            "/tmp/storage.csv.zst",
            bucket_name,
            "storage-usage/storage.csv.zst",
            ExtraArgs={"ACL": "public-read", "ContentType": "application/zstd"},
        )

    except ClientError as e:
        log.exception("Failed to upload storage.json.zst ", lException=e)

    try:
        s3_client.put_object(
            Bucket=bucket_name,
            Key="storage-usage/index.html",
            Body=INDEX_HTML,
            ContentType="text/html",
            ACL="public-read",
        )

    except Exception as e:
        log.exception("Failed to upload index.html", lException=e)
    log.info("Upload duration", f"{time.time() - start_time:.2f} secs")


def sendToOpenSearch(df: pd.DataFrame, hosts: str, http_auth: tuple[str, str]):
    """
    Sends the data to OpenSearch
    """
    log = gLogger.getSubLogger("sendToOpenSearch")
    log.info("Sending to OpenSearch")
    start_time = time.time()

    fields = {
        "SEName": {"type": "keyword"},
        "Name": {"type": "keyword"},
        "SESize": {"type": "long"},
        "SEFiles": {"type": "long"},
        "SESite": {"type": "keyword"},
        "SEType": {"type": "keyword"},
        "IsDisk": {"type": "boolean"},
        "production": {"type": "long"},
        "filetype": {"type": "keyword"},
        "maybe_eventtype": {"type": "long"},
        "online_stream": {"type": "keyword"},
        "ConfigName": {"type": "keyword"},
        "ConfigVersion": {"type": "keyword"},
        "Description": {"type": "keyword"},
        "ProcPath": {"type": "keyword"},
        "EventTypeID": {"type": "keyword"},
        "timestamp": {"type": "date"},
    }

    # Create an OpenSearch client
    client = OpenSearch(
        hosts=hosts,
        http_auth=http_auth,
        use_ssl=True,
        verify_certs=False,
        ssl_show_warn=False,
    )

    index_prefix = "lhcb_storageusage_index_"

    template_body = {
        "template": {"mappings": {"properties": fields}},
        "index_patterns": [f"{index_prefix}*"],
        "priority": 5,
    }
    result = client.indices.put_index_template(name=index_prefix, body=template_body)
    assert result["acknowledged"]

    from datetime import datetime, timezone

    utc_now = datetime.now(tz=timezone.utc)
    timestamp = utc_now.strftime("%Y-%m-%dT%H:%M")
    as_date = utc_now.isocalendar()
    index_timestamp = f"{as_date[0]}-{as_date[1]}"

    data = df.reset_index(drop=True).to_dict("records")

    actual_index = f"{index_prefix}{index_timestamp}"
    for entry in data:
        # entry.pop("Unnamed: 0")

        entry["_index"] = actual_index
        entry["_id"] = f"{entry['SEName']}_{entry['Name']}_{timestamp}"
        entry["timestamp"] = utc_now

    succeeded = []
    failed = []
    for success, item in helpers.parallel_bulk(
        client,
        actions=data,
        chunk_size=500,
        raise_on_error=False,
        raise_on_exception=False,
        max_chunk_bytes=20 * 1024 * 1024,
        request_timeout=60,
    ):

        if success:
            succeeded.append(item)
        else:
            failed.append(item)

    if len(failed) > 0:
        log.error("There were errors: ", f"{len(failed)}")
        for item in failed:
            log.error(f"    {item}")

    if len(succeeded) > 0:
        log.info("Bulk-inserted", f"{len(succeeded)} items.")
    log.info("Upload duration", f"{time.time() - start_time:.2f} secs")
    return S_OK()


class StorageUsageDumpAgent(AgentModule):

    def initialize(self):
        self.workDirectory = self.am_getWorkDirectory()
        self.s3_access_key = self.am_getOption("S3_access_key")
        self.s3_secret_key = self.am_getOption("S3_secret_key")
        self.s3_bucket_name = self.am_getOption("S3_bucket_name", S3_BUCKET_NAME)
        self.s3_endpoint_url = self.am_getOption("S3_endpoint_url", S3_ENDPOINT_URL)
        self.os_endpoint_url = self.am_getOption("OS_endpoint_url", OS_ENDPOINT_URL)
        self.os_username = gConfig.getValue("/Systems/NoSQLDatabases/User")
        self.os_password = gConfig.getValue("/Systems/NoSQLDatabases/Password")

        sql_db_params = returnValueOrRaise(getDBParameters("DataManagement/StorageUsageDB"))
        self.db_host = sql_db_params["Host"]
        self.db_port = sql_db_params["Port"]
        self.db_user = sql_db_params["User"]
        self.db_password = sql_db_params["Password"]
        self.db_name = sql_db_params["DBName"]

        # Sanity check: make sure that there are no duplicate event types
        if len(STREAM_TO_EVENTTYPE) != len(set(STREAM_TO_EVENTTYPE.values())):
            return S_ERROR("Duplicated event types !")

        return S_OK()

    def execute(self):
        asyncio.run(self.do_the_dump())

        return S_OK()

    async def do_the_dump(self):

        prod_data_info, bk_metadata = await asyncio.gather(
            get_prod_data_info(),
            get_bk_metadata(),
        )

        dfc_df = pd.DataFrame(prod_data_info[1], columns=prod_data_info[0])
        expected_total_size = dfc_df["SESize"].sum()
        expected_total_files = dfc_df["SEFiles"].sum()

        dfc_df["SESite"], dfc_df["SEType"] = zip(*dfc_df["SEName"].apply(se_name_to_type))
        dfc_df["IsDisk"] = ~dfc_df["SEType"].isin(TAPE_STORAGE_TYPES)

        # Find the bad rows
        bad_rows = dfc_df["Name"].str.contains(rf"^(?:{'|'.join(BAD_PREFIXES)})")

        # dfc_df_bad = dfc_df[bad_rows].reset_index(drop=True)
        # dfc_df = dfc_df[~bad_rows].reset_index(drop=True)
        dfc_df_bad = dfc_df[bad_rows]
        dfc_df = dfc_df[~bad_rows]

        # Dump the info about the bad rows
        for bad_prefix in BAD_PREFIXES:
            tmp_bad_rows = dfc_df_bad["Name"].str.startswith(bad_prefix)
            self.log.info(
                "Summary for bad prefix",
                (
                    f"{bad_prefix=}, rows: {np.count_nonzero(tmp_bad_rows)}, "
                    f"size: {dfc_df_bad.loc[tmp_bad_rows & (dfc_df_bad['SEName'] != 'FakeSE')]['SESize'].sum() / 1000**4}"
                ),
            )

        ### Treat the RAW data
        # /lhcb/data/<year>/RAW/<online stream>/<partition>/<config version>/<run number>
        # e.g. /lhcb/data/2016/RAW/FULL/LHCb/COLLISION16/173627

        raw_rows = dfc_df["Name"].str.startswith("/lhcb/data") & dfc_df["Name"].str.contains("/RAW/")

        # 173627 is the run number, -173627 is the production number
        dfc_df.loc[raw_rows, "production"] = -pd.to_numeric(dfc_df[raw_rows]["Name"].str.split("/", expand=True)[8])

        # In principle, this can only be RAW
        dfc_df.loc[raw_rows, "filetype"] = dfc_df.loc[raw_rows, "Name"].str.split("/", expand=True)[4]

        dfc_df.loc[raw_rows, "online_stream"] = dfc_df.loc[raw_rows, "Name"].str.split("/", expand=True)[5]

        # Add the event type name
        dfc_df.loc[raw_rows, "maybe_eventtype"] = dfc_df.loc[raw_rows, "online_stream"].apply(STREAM_TO_EVENTTYPE.get)

        # sanity: all raw files should have an event type
        if not dict(dfc_df.loc[raw_rows, "filetype"].value_counts()) == {"RAW": np.count_nonzero(raw_rows)}:
            raise NotImplementedError()

        ####### Treat the rows which are not RAW

        # e.g. /lhcb/MC/2017/PKTAUTAU.STRIP.DST/00230562/0000

        # 230562 is the production number
        dfc_df.loc[~raw_rows, "production"] = pd.to_numeric(dfc_df[~raw_rows]["Name"].str.split("/", expand=True)[5])

        dfc_df.loc[~raw_rows, "filetype"] = dfc_df.loc[~raw_rows, "Name"].str.split("/", expand=True)[4]

        bk_df = pd.DataFrame(bk_metadata[1], columns=bk_metadata[0])

        # Join the DFC info together with the BK info
        # dfc_df will map to multiple values in bk_df for real data due to different eventtypes (i.e. streams)
        df = dfc_df.merge(bk_df, on=["production", "filetype"], how="left", validate="many_to_many")

        # Replace Nan with 0.0 or Unknown

        group_cols = list(bk_df.columns)
        group_cols.pop(group_cols.index("production"))

        # The original version was filtering on ConfigName being Nan,  but we get the same result whn doing it no matter what
        # df.loc[pd.isnull(df["ConfigName"])] = df.loc[pd.isnull(df["ConfigName"])].fillna(
        #     {k: "Unknown" if df.dtypes[k].type == np.object_ else 0 for k in group_cols}
        # )
        df.fillna({k: "Unknown" if df.dtypes[k].type == np.object_ else 0 for k in group_cols}, inplace=True)

        df["EventTypeID"] = df["EventTypeID"].astype(int)

        raw_rows = df["Name"].str.contains("/RAW/")

        # In dfc_df ["SEName", "Name"] is unique, but when joining
        # on bk_df we made duplicates due to real data having
        # multiple event types (i.e. streams) for a given
        # production (i.e. -run number) and filetype (e.g. RAW)
        df_tmp = (
            df.loc[raw_rows]
            .groupby(["SEName", "Name"])
            .apply(find_matching_event_type_row, include_groups=False)
            .reset_index()
        )

        df = pd.concat([df_tmp, df.loc[~raw_rows]]).reset_index(drop=True)

        df = pd.concat([df, dfc_df_bad])

        # Redo it because dfc_df_bad wasn't there then
        df.fillna(
            {
                k: "Unknown" if df.dtypes[k].type == np.object_ else 0
                for k in group_cols + ["online_stream", "maybe_eventtype", "Description", "production"]
            },
            inplace=True,
        )

        # This somehow explodes the memory....

        # for cat in ("SEName", "SESite", "SEType", "online_stream", "ConfigName"):
        #     df[cat] = df[cat].astype("category")
        # for cat in ("SESize", "SEFiles", "EventTypeID"):
        #     df[cat] = pd.to_numeric(df[cat], downcast="unsigned")

        # for cat in ("production",):
        #     df[cat] = df[cat].astype(int)

        if len(df) != len(df[["SEName", "Name"]].drop_duplicates()):
            raise NotImplementedError("Something is very wrong")
        if expected_total_files != df["SEFiles"].sum():
            raise NotImplementedError()
        if expected_total_size != df["SESize"].sum():
            raise NotImplementedError()

        df.to_csv("/tmp/storage.csv.zst", index=False)
        result = {
            "groupings": [
                {
                    "name": "Logical vs Disk vs Tape",
                    "groups": [
                        {
                            "title": "Logical space",
                            "id": "logical",
                            "data": summarise_space(df[df["SEName"] == "FakeSE"], group_cols),
                        },
                        {
                            "title": "Physical space - Disk",
                            "id": "physical-disk",
                            "data": summarise_space(df[(df["SEName"] != "FakeSE") & df["IsDisk"]], group_cols),
                        },
                        {
                            "title": "Physical space - Tape",
                            "id": "physical-tape",
                            "data": summarise_space(df[(df["SEName"] != "FakeSE") & ~df["IsDisk"]], group_cols),
                        },
                        {
                            "title": "Physical space",
                            "id": "physical",
                            "data": summarise_space(df[df["SEName"] != "FakeSE"], group_cols),
                        },
                    ],
                },
                {
                    "name": "Physical by SE type",
                    "groups": [
                        {
                            "title": se_type,
                            "id": se_type.lower(),
                            "data": summarise_space(
                                df[(df["SEName"] != "FakeSE") & (df["SEType"] == se_type)], group_cols
                            ),
                        }
                        for se_type in df["SEType"].unique()
                    ],
                },
                {
                    "name": "Physical by SE site (disk)",
                    "groups": [
                        {
                            "title": se_site,
                            "id": se_site.lower(),
                            "data": summarise_space(
                                df[df["IsDisk"] & (df["SEName"] != "FakeSE") & (df["SESite"] == se_site)], group_cols
                            ),
                        }
                        for se_site in sorted(
                            df["SESite"].unique(), key=lambda x: df[df["SESite"] == x]["SESize"].sum(), reverse=True
                        )
                        if se_site != "FakeSE"
                    ],
                },
                {
                    "name": "Physical by SE site (tape)",
                    "groups": [
                        {
                            "title": se_site,
                            "id": se_site.lower(),
                            "data": summarise_space(
                                df[~df["IsDisk"] & (df["SEName"] != "FakeSE") & (df["SESite"] == se_site)], group_cols
                            ),
                        }
                        for se_site in sorted(
                            df["SESite"].unique(), key=lambda x: df[df["SESite"] == x]["SESize"].sum(), reverse=True
                        )
                        if se_site != "FakeSE"
                    ],
                },
            ],
            "last-updated": datetime.now(tz=timezone.utc).isoformat(),
        }

        path = Path("/tmp/storage.json.zst")
        path.write_bytes(zstandard.compress(json.dumps(result).encode()))

        uploadToS3(
            endpoint_url=self.s3_endpoint_url,
            bucket_name="lhcbdirac",
            access_key=self.s3_access_key,
            secret_key=self.s3_secret_key,
        )

        engine = create_engine(
            f"mysql://{self.db_user}:{self.db_password}@{self.db_host}:{self.db_port}/{self.db_name}",
            echo=False,
        )
        Base.metadata.create_all(engine)

        # df.to_csv("/home/dirac/df.csv")
        await asyncio.gather(
            returnValueOrRaiseAsync(
                partial(sendToOpenSearch, df, self.os_endpoint_url, (self.os_username, self.os_password))
            ),
            returnValueOrRaiseAsync(
                partial(
                    directory_metadata_to_sql,
                    df,
                    db_host=self.db_host,
                    db_name=self.db_name,
                    db_user=self.db_user,
                    db_password=self.db_password,
                    db_port=self.db_port,
                    engine=engine,
                )
            ),
            returnValueOrRaiseAsync(
                partial(
                    storage_summary_to_sql,
                    df,
                    db_host=self.db_host,
                    db_name=self.db_name,
                    db_user=self.db_user,
                    db_password=self.db_password,
                    db_port=self.db_port,
                    engine=engine,
                )
            ),
        )


INDEX_HTML = """
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title>Storage Usage Visualization</title>
  <style>
    body {
      font-family: Arial, sans-serif;
      margin: 0;
      background-color: #1e1e1e;
      color: #e0e0e0;
    }
    h1, h2 {
      color: #e0e0e0;
      text-align: center;
    }
    h1 {
      margin: 0;
    }
    a {
      color: #8888ff;
    }
    #heading {
      width: 100%;
      position: sticky;
      top: 0;
      background-color: #333333;
      color: #e0e0e0;
      padding: 10px 0;
    }
    #heading #container {
      display: flex;
      flex-direction: column;
      align-items: center;
    }
    #breadcrumbs {
      margin: 10px;
      text-align: center;
    }
    #breadcrumbs span + span::before {
      content: " / ";
      color: #e0e0e0;
      margin: 0 5px;
    }
    #reset {
      margin: 10px;
      display: block;
      margin-left: auto;
      margin-right: auto;
      background-color: #555555;
      color: #ffffff;
      border: 1px solid #888888;
    }
    select {
      background-color: #555555;
      color: #ffffff;
    }
    .container {
      display: flex;
      flex-wrap: wrap;
      justify-content: space-around;
      width: 100%;
    }
    .tooltip {
      position: absolute;
      visibility: hidden;
      padding: 8px;
      border-radius: 4px;
      background: #444444;
      color: #ffffff;
    }

    /* Modal styles */
    #loading-modal {
      display: flex;
      justify-content: center;
      align-items: center;
      position: fixed;
      top: 0;
      left: 0;
      width: 100%;
      height: 100%;
      background-color: rgba(0, 0, 0, 0.5);
      z-index: 1000;
      visibility: visible;
      opacity: 1;
      transition: opacity 0.5s ease, visibility 0.5s ease;
    }
    #loading-modal.hidden {
      visibility: hidden;
      opacity: 0;
    }
    #loading-content {
      background-color: #333333;
      color: #ffffff;
      display: flex;
      justify-content: center;
      align-items: center;
      padding: 20px;
      border-radius: 8px;
      text-align: center;
      box-shadow: 0 4px 8px rgba(0, 0, 0, 0.2);
    }
    .spinner {
      border-top: 4px solid #8888ff;
      border: 4px solid rgba(255, 255, 255, 0.1);
      border-radius: 50%;
      margin: 1em;
      width: 40px;
      height: 40px;
      animation: spin 1s linear infinite;
    }
    .view-option {
      display: inline-block;
      margin: 0 10px;
    }
    @keyframes spin {
      0% { transform: rotate(0deg); }
      100% { transform: rotate(360deg); }
    }
  </style>
</head>
<body>
  <div id="loading-modal">
    <div id="loading-content">
      <div class="spinner"></div>
      <p>Loading, please wait...</p>
    </div>
  </div>

  <div id="heading">
    <h1>Storage Usage Visualization</h1>
    <div id="breadcrumbs"></div>
    <button id="reset">Reset breadcrumbs</button>
    <div style="text-align: center; margin-top: 10px;">
      <div class="view-option">
        <label for="grouping-select">Choose split:</label>
        <select id="grouping-select"></select>
      </div>
      <div class="view-option">
        <label for="hide-conditions">Hide conditions:</label>
        <input type="checkbox" id="hide-conditions">
      </div>
    </div>
    <div style="text-align: center;">
      Mouse over a segment for details. Click to zoom in.
      Last updated: <span id="last-updated-time"></span>
    </div>
  </div>
  <div id="charts" class="container">
  </div>
  <div id="tooltip" class="tooltip"></div>

  <script src="https://cdn.jsdelivr.net/npm/fzstd@0.1.1/umd/index.js"></script>
  <script src="https://d3js.org/d3.v6.min.js"></script>
  <script src="script.js"></script>
</body>
</html>
"""


def directory_metadata_to_sql(
    df, *, db_user: str, db_password: str, db_host: str, db_port: str, db_name: str, engine: Engine
):
    log = gLogger.getLocalSubLogger("directory_metadata_to_sql")
    log.info("Writing directory metadata to database")
    start_time = time.time()

    df = df[df["SEName"] == "FakeSE"][
        [
            "Name",
            "production",
            "filetype",
            "online_stream",
            "maybe_eventtype",
            "ConfigName",
            "ConfigVersion",
            "Description",
            "ProcPath",
            "EventTypeID",
        ]
    ].drop_duplicates()

    # Create the database and table
    # engine = create_engine(
    #     f"mysql://{db_user}:{db_password}@{db_host}:{db_port}/{db_name}",
    #     echo=False,
    # )
    # Base.metadata.create_all(engine)

    # Assuming 'engine' is your database connection
    with engine.connect() as conn:
        conn.execute(delete(DirectoryMetadata))
        # Base.metadata.drop_all(engine, tables=[DirectoryMetadata.__table__])
        # conn.execute(text(f"TRUNCATE TABLE {DirectoryMetadata.__tablename__}"))
        df.to_sql(
            DirectoryMetadata.__tablename__,
            conn,
            index=False,
            method="multi",
            chunksize=10000,
            if_exists="append",
        )
        new_comment = json.dumps({"LastUpdate": datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")})
        conn.execute(text(f"ALTER TABLE {DirectoryMetadata.__tablename__} COMMENT = '{new_comment}';"))

    log.info("Upload duration", f"{time.time() - start_time:.2f} secs")
    return S_OK()


def storage_summary_to_sql(
    df, *, db_user: str, db_password: str, db_host: str, db_port: str, db_name: str, engine: Engine
):
    log = gLogger.getLocalSubLogger("storage_summary_to_sql")
    log.info("Writing storage summary to database")
    start_time = time.time()

    df = df[
        [
            "SEName",
            "SESize",
            "SEFiles",
            "SESite",
            "SEType",
            "IsDisk",
        ]
    ]

    df_size = df[["SEName", "SESize", "SEFiles"]].groupby("SEName").sum()
    df_meta = df[["SEName", "SESite", "IsDisk", "SEType"]].drop_duplicates().set_index("SEName")
    df_summary = df_size.merge(df_meta, on="SEName").reset_index()

    # # Create the database and table
    # engine = create_engine(
    #     f"mysql://{db_user}:{db_password}@{db_host}:{db_port}/{db_name}",
    #     echo=False,
    # )
    # Base.metadata.create_all(engine)

    # Assuming 'engine' is your database connection
    with engine.connect() as conn:
        conn.execute(delete(StorageSummary))
        df_summary.to_sql(
            StorageSummary.__tablename__,
            conn,
            index=False,
            method="multi",
            chunksize=10000,
            if_exists="append",
        )
        new_comment = json.dumps({"LastUpdate": datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")})
        conn.execute(text(f"ALTER TABLE {StorageSummary.__tablename__} COMMENT = '{new_comment}';"))

    log.info("Upload duration", f"{time.time() - start_time:.2f} secs")
    return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit test for the MCExtensionAgent."""
import datetime
import copy
import importlib
import unittest

from unittest.mock import Mock, MagicMock
from DIRAC import S_OK, S_ERROR, gLogger

from LHCbDIRAC.TransformationSystem.Agent.MCExtensionAgent import MCExtensionAgent


class MCExtensionAgentTestCase(unittest.TestCase):
    def setUp(self):
        self.mockAM = MagicMock()
        self.mockAM.am_getOption.return_value = S_OK()
        self.agent = importlib.import_module("LHCbDIRAC.TransformationSystem.Agent.MCExtensionAgent")
        self.agent.AgentModule = self.mockAM
        self.agent.DIRACMCExtensionAgent = self.mockAM
        self.agent = MCExtensionAgent()
        print(dir(self.agent))
        self.agent.log = gLogger

    # def test_execute( self ):
    #   # self.agent = self.moduleTested.MCExtensionAgent( 'MCExtensionAgent', 'MCExtensionAgent', 'MCExtensionAgent' )
    #   self.agent.transformationTypes = ['MCSimulation', 'Simulation']
    #   self.agent.rpcProductionRequest = Mock()
    #   self.agent._checkProductionRequest = Mock()
    #
    #   ###########################################################################
    #
    #   # disabled by configuration
    #   self.mockAM.am_getOption.side_effect = lambda optionName,
    # defaultValue: {( 'EnableFlag', 'True' ) : 'False'}[( optionName, defaultValue )]
    #
    #   ret = self.agent.execute()
    #   self.assertTrue( ret['OK'] )
    #
    #   ###########################################################################
    #
    #   # bad request
    #   # self.mockAM.am_getOption.side_effect = lambda optionName,
    # defaultValue: {( 'EnableFlag', 'True' ) : 'True'}[( optionName, defaultValue )]
    #   self.agent.rpcProductionRequest.getProductionRequestSummary.return_value = S_ERROR()
    #
    #   ret = self.agent.execute()
    #   self.assertFalse( ret['OK'] )
    #
    #   ###########################################################################
    #
    #   # normal operation
    #   # self.mockAM.am_getOption.side_effect = lambda optionName,
    # defaultValue: {( 'EnableFlag', 'True' ) : 'True'}[( optionName, defaultValue )]
    #   productionRequests = {10964: {'bkTotal': 259499L, 'master': 10960, 'reqTotal': 500000L},
    #                         10965: {'bkTotal': 610999L, 'master': 10960, 'reqTotal': 1000000L},
    #                         10966: {'bkTotal': 660995L, 'master': 10959, 'reqTotal': 1000000L}}
    #   self.agent.rpcProductionRequest.getProductionRequestSummary.return_value = S_OK( productionRequests )
    #   self.agent._checkProductionRequest.return_value = S_OK()
    #
    #   ret = self.agent.execute()
    #   self.assertTrue( ret['OK'] )
    #   self.assertEqual( self.agent._checkProductionRequest.call_count, len( productionRequests ) )
    #   self.agent._checkProductionRequest.assert_has_calls(
    # [call( ID, summary ) for ID, summary in productionRequests.items()] )

    def test__checkProductionRequest(self):
        # self.agent = self.moduleTested.MCExtensionAgent( 'MCExtensionAgent', 'MCExtensionAgent', 'MCExtensionAgent' )
        self.agent.transformationTypes = ["MCSimulation", "Simulation"]
        self.agent.rpcProductionRequest = Mock()
        self.agent.transClient = Mock()
        self.agent._extendProduction = Mock()

        completeProductionRequestID = 12417
        completeProductionRequestSummary = {"bkTotal": 503999, "master": 12415, "reqTotal": 500000}

        incompleteProductionRequestID = 11162
        incompleteProductionRequestSummary = {"bkTotal": 296999, "master": 11158, "reqTotal": 500000}
        missingEventsExp = 203001

        productionsProgressNoStripping = {
            "Rows": [
                {"BkEvents": 390499, "ProductionID": 24781, "RequestID": 11162, "Used": 0},
                {"BkEvents": 296999, "ProductionID": 24782, "RequestID": 11162, "Used": 1},
            ]
        }
        productionsProgressWithStripping = {
            "Rows": [
                {"BkEvents": 2969990, "ProductionID": 24781, "RequestID": 11162, "Used": 0},
                {"BkEvents": 296999, "ProductionID": 24782, "RequestID": 11162, "Used": 1},
            ]
        }
        extensionFactorExp = 10.0

        transformations = {
            24781: {
                "AgentType": "Automatic",
                "AuthorDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=romanov/CN=427293/CN=Vladimir Romanovskiy",
                "AuthorGroup": "lhcb_prod",
                "Body": "<Workflow>...</Workflow>",
                "CreationDate": datetime.datetime(2013, 6, 10, 18, 30, 22),
                "Description": "",
                "EventsPerTask": 0,
                "FileMask": "",
                "GroupSize": 1,
                "InheritedFrom": 0,
                "LastUpdate": datetime.datetime(2013, 6, 12, 6, 59, 56),
                "LongDescription": "prodDescription",
                "MaxNumberOfTasks": 0,
                "Plugin": "",
                "Status": "Idle",
                "TransformationFamily": "11158",
                "TransformationGroup": "Sim08a/Digi13/Trig0x409f0045/Reco14a/Stripping20NoPrescalingFlag",
                "TransformationID": 24781,
                "TransformationName": "Request_11162_bof.xml",
                "Type": "MCSimulation",
            },
            24782: {
                "AgentType": "Automatic",
                "AuthorDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=romanov/CN=427293/CN=Vladimir Romanovskiy",
                "AuthorGroup": "lhcb_prod",
                "Body": "<Workflow>...<Workflow>",
                "CreationDate": datetime.datetime(2013, 6, 10, 18, 30, 27),
                "Description": "",
                "EventsPerTask": 0,
                "FileMask": "",
                "GroupSize": 5,
                "InheritedFrom": 0,
                "LastUpdate": datetime.datetime(2013, 6, 12, 6, 59, 56),
                "LongDescription": "prodDescription",
                "MaxNumberOfTasks": 0,
                "Plugin": "BySize",
                "Status": "Idle",
                "TransformationFamily": "11158",
                "TransformationGroup": "Sim08a/Digi13/Trig0x409f0045/Reco14a/Stripping20NoPrescalingFlag",
                "TransformationID": 24782,
                "TransformationName": "Request_bof.xml",
                "Type": "Merge",
            },
        }
        simulation = transformations[24781]

        ###########################################################################

        # complete production request, no extension needed
        ret = self.agent._checkProductionRequest(completeProductionRequestID, completeProductionRequestSummary)
        self.assertTrue(ret["OK"])
        self.assertFalse(self.agent._extendProduction.called)
        self.agent._extendProduction.reset_mock()

        ###########################################################################

        # failed request for production progress
        self.agent.rpcProductionRequest.getProductionProgressList.return_value = S_ERROR()

        ret = self.agent._checkProductionRequest(incompleteProductionRequestID, incompleteProductionRequestSummary)
        self.assertFalse(ret["OK"])
        self.assertFalse(self.agent._extendProduction.called)
        self.agent._extendProduction.reset_mock()

        ###########################################################################

        # failed request for production
        self.agent.rpcProductionRequest.getProductionProgressList.return_value = S_OK(productionsProgressNoStripping)
        self.agent.transClient.getTransformation.return_value = S_ERROR()

        ret = self.agent._checkProductionRequest(incompleteProductionRequestID, incompleteProductionRequestSummary)
        self.assertFalse(ret["OK"])
        self.assertFalse(self.agent._extendProduction.called)
        self.agent._extendProduction.reset_mock()

        ###########################################################################

        # simulation production not found
        self.agent.rpcProductionRequest.getProductionProgressList.return_value = S_OK(productionsProgressNoStripping)
        noSimulationTransformations = copy.deepcopy(transformations)
        for t in noSimulationTransformations.values():
            t["Type"] = "Merge"
        self.agent.transClient.getTransformation.side_effect = lambda transformationID: S_OK(
            noSimulationTransformations[transformationID]
        )

        ret = self.agent._checkProductionRequest(incompleteProductionRequestID, incompleteProductionRequestSummary)
        self.assertFalse(ret["OK"])
        self.assertFalse(self.agent._extendProduction.called)
        self.agent._extendProduction.reset_mock()

        ###########################################################################

        # non idle simulation
        self.agent.rpcProductionRequest.getProductionProgressList.return_value = S_OK(productionsProgressNoStripping)
        nonIdleSimulationTransformations = copy.deepcopy(transformations)
        for t in nonIdleSimulationTransformations.values():
            if t["Type"] == "MCSimulation":
                t["Status"] = "Active"
        self.agent.transClient.getTransformation.side_effect = lambda transformationID: S_OK(
            nonIdleSimulationTransformations[transformationID]
        )

        ret = self.agent._checkProductionRequest(incompleteProductionRequestID, incompleteProductionRequestSummary)
        self.assertTrue(ret["OK"])
        self.assertFalse(self.agent._extendProduction.called)
        self.agent._extendProduction.reset_mock()

        ###########################################################################

        # no stripping production (no extension factor)
        self.agent.rpcProductionRequest.getProductionProgressList.return_value = S_OK(productionsProgressNoStripping)
        self.agent.transClient.getTransformation.side_effect = lambda transformationID: S_OK(
            transformations[transformationID]
        )
        self.agent._extendProduction.return_value = S_OK()

        #     ret = self.agent._checkProductionRequest( incompleteProductionRequestID, incompleteProductionRequestSummary )
        #     self.assertTrue( ret['OK'] )
        #     self.agent._extendProduction.assert_called_once_with( simulation, 1.0, missingEventsExp )
        #     self.agent._extendProduction.reset_mock()

        ###########################################################################

        # stripping production (extension factor)
        self.agent.rpcProductionRequest.getProductionProgressList.return_value = S_OK(productionsProgressWithStripping)
        self.agent.transClient.getTransformation.side_effect = lambda transformationID: S_OK(
            transformations[transformationID]
        )
        self.agent._extendProduction.return_value = S_OK()

    #     ret = self.agent._checkProductionRequest( incompleteProductionRequestID, incompleteProductionRequestSummary )
    #     self.assertTrue( ret['OK'] )
    #     self.agent._extendProduction.assert_called_once_with( simulation, extensionFactorExp, missingEventsExp )
    #     self.agent._extendProduction.reset_mock()

    def test__extendProduction(self):
        # self.agent = self.moduleTested.MCExtensionAgent( 'MCExtensionAgent', 'MCExtensionAgent', 'MCExtensionAgent' )
        self.agent.transformationTypes = ["MCSimulation", "Simulation"]
        self.agent.transClient = Mock()
        self.agent.transClient.extendTransformation.return_value = S_OK()
        self.agent.transClient.setTransformationParameter.return_value = S_OK()

        self.agent.cpuTimeAvg = 1000000
        self.agent.cpuNormalizationFactorAvg = 1.0

        production = {
            "AgentType": "Automatic",
            "AuthorDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=romanov/CN=427293/CN=Vladimir Romanovskiy",
            "AuthorGroup": "lhcb_prod",
            "Body": """
        <Workflow>
          <origin></origin>
          <description><![CDATA[prodDescription]]></description>
          <descr_short></descr_short>
          <version>0.0</version>
          <type></type>
          <name>Request_12416_MCSimulation_Sim08a/Digi13/Trig0x40760037/Reco14a/Stripping20bof</name>
          <Parameter name="CPUe" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="C">
            <value><![CDATA[10]]></value>
          </Parameter>
        </Workflow>
      """,
            "CreationDate": datetime.datetime(2013, 6, 4, 8, 12),
            "Description": "",
            "EventsPerTask": 0,
            "FileMask": "",
            "GroupSize": 1,
            "InheritedFrom": 0,
            "LastUpdate": datetime.datetime(2013, 6, 4, 8, 12, 3),
            "LongDescription": "prodDescription",
            "MaxNumberOfTasks": 0,
            "Plugin": "",
            "Status": "Idle",
            "TransformationFamily": "12415",
            "TransformationGroup": "Sim08a/Digi13/Trig0x40760037/Reco14a/Stripping20r1NoPrescalingFl",
            "TransformationID": 24614,
            "TransformationName": "Request_12416_bof.xml",
            "Type": "MCSimulation",
        }
        extensionFactor = 2.0
        eventsNeeded = 1000000

        productionIDExp = 24614
        numberOfTasksExp = 32

        ret = self.agent._extendProduction(production, extensionFactor, eventsNeeded)
        self.assertTrue(ret["OK"])
        self.agent.transClient.extendTransformation.assert_called_once_with(productionIDExp, numberOfTasksExp)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(MCExtensionAgentTestCase)
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: UserStorageUsageAgent

.. module: UserStorageUsageAgent

:synopsis: UserStorageUsageAgent simply inherits the StorageUsage agent
  and loops over the /lhcb/user directory, removing empty ones.
"""
# # imports
from DIRAC import S_OK
from LHCbDIRAC.DataManagementSystem.Agent.StorageUsageAgent import StorageUsageAgent
from DIRAC.Core.Utilities import List


class UserStorageUsageAgent(StorageUsageAgent):
    """
    .. class:: UserStorageUsageAgent

    """

    def removeEmptyDir(self, dirPath):
        """remove empty directories, but skip home.

        :param self: self reference
        :param str dirPath: directory to remove
        """
        # Do not remove user's home dir
        if len(List.fromChar(dirPath, "/")) > 4:
            return StorageUsageAgent.removeEmptyDir(self, dirPath)
        return S_OK()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Upload a file to the grid storage and register it in the File Catalog."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... LFN Path SE [GUID]",
                "Arguments:",
                "  LFN:      Logical File Name",
                "  Path:     Local path of the file",
                "  SE:       DIRAC Storage Element",
                "  GUID:     GUID to use in the registration (optional)",
                "",
                " ** OR **",
                "",
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... LocalFile",
                "Arguments:",
                "  LocalFile: Path to local file containing all the above, i.e.:",
                "  lfn1 localfile1 SE [GUID1]",
                "  lfn2 localfile2 SE [GUID2]",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeAddFile
    from DIRAC import exit

    exit(executeAddFile())


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Retrieve and verify information about the inputs to an Analysis Production
"""
from __future__ import annotations

import json
from datetime import datetime, timezone
import re
from itertools import islice
from pathlib import Path
import sys

from rich.progress import track

from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise, convertToReturnValue
from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient
from LHCbDIRAC.ProductionManagementSystem.Client.AnalysisProductionsClient import AnalysisProductionsClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


def parseArgs():
    version = ""
    with_ancestors = False
    output = ""

    @convertToReturnValue
    def setVersion(s: str):
        nonlocal version
        version = s

    @convertToReturnValue
    def withAncestors(_):
        nonlocal with_ancestors
        with_ancestors = True

    @convertToReturnValue
    def setOutput(s: str):
        nonlocal output
        output = s

    switches = [
        ("", "version=", "Version of the Analysis Production (optional)", setVersion),
        (
            "",
            "with-ancestors",
            "Return all ancestors of the production -- usually a LOT of files. Not recommended unless --output is also set.",
            withAncestors,
        ),
        ("", "output=", "Path to a .json file to dump the output to", setOutput),
    ]
    Script.registerSwitches(switches)
    Script.registerArgument("wg: Name of the WG the production belongs to e.g. `bnoc`")
    Script.registerArgument("analysis: Name of the analysis e.g. `bds2kstkstb`")
    Script.registerArgument("name: Name of the job in the production e.g. `13104007_bs2kst_892_kstb_892_2011_magup`")
    Script.parseCommandLine(ignoreErrors=False)
    wg, analysis, name = Script.getPositionalArgs(group=True)
    wg = wg.lower()
    analysis = analysis.lower()
    name = name.lower()

    if not ConfigurationClient().ping()["OK"]:
        gLogger.fatal("Failed to contact CS, do you have a valid proxy?")
        sys.exit(1)

    return (
        wg,
        analysis,
        name,
        version,
        with_ancestors,
        output,
    )


@Script()
def main():
    wg, analysis, name, version, with_ancestors, output = parseArgs()
    config = {
        "wg": wg,
        "analysis": analysis,
        "name": name,
    }
    if version != "":
        config["version"] = version
    save_output = False
    if output != "":
        output = Path(output)
        save_output = True

    tc = TransformationClient()
    bk = BookkeepingClient()
    apc = AnalysisProductionsClient()

    apd_info = _get_info_from_apd(apc, config)
    # Add the ancestors information
    all_ancestors = _get_ancestors_info(apd_info, bk)
    relational_ancestors = [{"lfn": lfn, "ancestors": []} for lfn in apd_info["apd_lfns"]]

    relational_ancestors = _attach_ancestors(relational_ancestors, all_ancestors, 1, apd_info["depth"])

    # Check the ancestors info matches the bk query and no duplicate files
    extra_info = _check_inputs(apd_info, all_ancestors, bk, tc)
    oldest_ancestors = []
    # pylint: disable-next=E1126
    for anc_list in all_ancestors[apd_info["depth"]].values():
        oldest_ancestors += anc_list

    events_ancestors, empty_files_ancestors = _get_total_num_events(oldest_ancestors, bk)
    events_expected, empty_files_expected = _get_total_num_events(extra_info["input_lfns"], bk)

    if events_ancestors != events_expected:
        gLogger.warn("The number of input events does not match the expeced number of events from bookeeping")
        gLogger.warn("Check `missing_ancestors` or `extra_ancestors` to see if a file was skipped or duplicated.")
    gLogger.always(f"Input events: {events_ancestors}")
    gLogger.always(f"Expected events: {events_expected}")

    all_info = {
        "ancestors_timestamp": datetime.now(tz=timezone.utc).isoformat(),
        "wg": config["wg"],
        "analysis": config["analysis"],
        "version": _get_aprod_version(apd_info, config),
        "name": config["name"],
        "input_query": extra_info["query"],
        "n_input_events": events_ancestors,
        "n_expected_events": events_expected,
        "missing_ancestors": extra_info["missing_ancestors"],
        "extra_ancestors": extra_info["extra_ancestors"],
        "duplicate_ancestors": extra_info["duplicate_ancestors"],
    }

    if with_ancestors:
        all_info["ancestors"] = relational_ancestors

    if save_output:
        with open(output, "w") as out_f:
            gLogger.info(f"Writing info to requested json file {output}")
            json.dump(all_info, out_f, indent=4)
    else:
        # could maybe make the output a bit nicer
        # i.e. some fields only on verbose
        for info_name, field in all_info.items():
            if info_name != "ancestors":
                gLogger.info(f"{info_name}: {field}")
        if with_ancestors:
            gLogger.warn("All ancestors information only prints to stdout if verbose (`use -d`)")
            gLogger.verbose("Ancestors:\n")
            gLogger.verbose(all_info["ancestors"])


def _batched(iterable, n):
    """
    batched("ABCDEFG", 3) -> ABC DEF G
    """
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


def _get_info_from_apd(apc, apc_kwargs):
    apd_info = {"apd_lfns": [], "depth": None}
    prods = returnValueOrRaise(
        apc.getProductions(**apc_kwargs, with_transformations=True, with_lfns=True, with_pfns=False)
    )

    if len(prods) == 0:
        raise ValueError(f"Could not find a production matching tags {apc_kwargs}")
    if len(prods) > 1:
        raise NotImplementedError(f"Tags must match only one production")

    prod = prods[0]

    # Output lfns from this production
    lfns = prod["lfns"]
    apd_info["apd_lfns"] = lfns

    # Should always be 2 but can have > 2 steps
    if len(prod["transformations"]) != 2:
        raise NotImplementedError("Production has more than 2 transformations!")

    depth = sum([len(x["steps"]) for x in prod["transformations"]])
    apd_info["depth"] = depth

    gLogger.always(f"{len(lfns)} LFNs and depth of {depth}")
    apd_info["transformations"] = prod["transformations"]
    return apd_info


def _get_aprod_version(apd_info, config):
    """
    Version is an optional argument but if not supplied can be determined
    from the apd_info
    """
    if "version" not in config:
        found_version = False
        for transformation in apd_info["transformations"]:
            for step in transformation["steps"]:
                if "extras" in step:
                    version = re.findall(
                        r"AnalysisProductions\.(v\d+r\d+p\d+|v\dr\d+)",
                        " ".join(step["extras"]),
                    )
                    if len(version) == 1:
                        return version[0]
                    elif len(version) > 1:
                        raise NotImplementedError(
                            f"Found multiple versions in production:\
                            {step['extras']} for transformation {transformation}"
                        )
        if not found_version:
            raise RuntimeError(f"Could not find AnalysisProduction version from {apd_info}")
    else:
        version = config["version"]
    return version


def _get_ancestors_info(apd_info, bk):
    """
    Get the ancestors for every LFN produced by the Aprod
    A separate step `attach_ancestors` will then recover the relationship
    between the different levels of LFNs.
    """
    all_ancestors = [apd_info["apd_lfns"]]
    for depth in range(apd_info["depth"]):
        if depth == 0:
            lfn_names = all_ancestors[depth]
        else:
            lfn_names = []
            for anc_dict in all_ancestors[depth].values():
                lfn_names += [parent_dict["lfn"] for parent_dict in anc_dict]
        ancestors = {}
        for lfn_batch in track(
            _batched(lfn_names, 50), total=len(lfn_names) // 50, description=f"Getting ancestors at depth {depth + 1}"
        ):
            ancestors_dict = returnValueOrRaise(bk.getFileAncestors(list(lfn_batch), depth=1, replica=False))
            if ancestors_dict["Failed"] != []:
                raise RuntimeError(f"Failed to get all file ancestors: {ancestors_dict['Failed']}")
            ancestors |= ancestors_dict["Successful"]
        # need to use getFileMetadata() as well to get the size
        ancestors_trimmed = {}
        lfns = []
        for child_lfn, ancestor_dicts in ancestors.items():
            lfns += [[anc_dict["FileName"], child_lfn] for anc_dict in ancestor_dicts]

        for lfn_batch in track(
            _batched(lfns, 10_000), total=len(lfns) // 10_000, description=f"Getting file metadata at depth {depth + 1}"
        ):
            childs = []
            parents = []
            for lfn_names in lfn_batch:
                parent, child = lfn_names
                childs.append(child)
                parents.append(parent)
            metadata = returnValueOrRaise(bk.getFileMetadata(parents))
            if metadata["Failed"] != []:
                raise RuntimeError(f"Failed to get metadata for {metadata['Failed']}")
            meta_success = metadata["Successful"]
            for child, parent in zip(childs, parents):
                if child not in ancestors_trimmed:
                    ancestors_trimmed[child] = []
                ancestors_trimmed[child] += [
                    {
                        "lfn": parent,
                        "n_events": meta_success[parent]["EventStat"],
                        "size": meta_success[parent]["FileSize"],
                    }
                ]
        all_ancestors.append(ancestors_trimmed)

    return all_ancestors


def _attach_ancestors(top_dict, all_dicts, i, depth):
    """
    Recursively attach the ancestor LFNs from each transformation
    to the resulting LFN
    """
    for j, lfn_dict in enumerate(top_dict):
        lfn_dict["ancestors"] = all_dicts[i][lfn_dict["lfn"]]
        if i + 1 <= depth:
            lfn_dict["ancestors"] = _attach_ancestors(lfn_dict["ancestors"], all_dicts, i + 1, depth)
            top_dict[j] = lfn_dict
    return top_dict


def _check_inputs(apd_info, all_ancestors, bk, tc):
    """
    Checks the ancestors against the bk query result
    Should hopefully be the same but in case they are not we can report it
    """
    # These will always be the deepest LFNs
    from_ancestors_lfns = []
    for anc_list in all_ancestors[apd_info["depth"]].values():
        from_ancestors_lfns += [anc_dict["lfn"] for anc_dict in anc_list]
    output_info = {"missing_ancestors": [], "extra_ancestors": [], "from_ancestors_lfns": from_ancestors_lfns}
    # Already check earlier we should have only 2 transformations but check
    # which is the merge step
    prev_tid = apd_info["transformations"][0]["id"]
    next_tid = apd_info["transformations"][1]["id"]
    query = None
    prev_query = returnValueOrRaise(tc.getBookkeepingQuery(prev_tid))
    next_query = returnValueOrRaise(tc.getBookkeepingQuery(next_tid))
    if "ProductionID" in prev_query and prev_query["ProductionID"] == next_tid:
        query = next_query
    elif "ProductionID" in next_query and next_query["ProductionID"] == prev_tid:
        query = prev_query
    else:
        raise RuntimeError(
            f"Could not determine the merge step for\
                           {apd_info['transformations']}"
        )

    input_lfns = returnValueOrRaise(bk.getFiles(query))
    output_info["input_lfns"] = input_lfns
    output_info["query"] = query
    found_mismatch = False
    if len(input_lfns) != len(from_ancestors_lfns):
        gLogger.warn(
            "Mismatch between number of input LFNs from bookeeping query"
            + f" ({len(input_lfns)}) and ancestor information ({len(from_ancestors_lfns)})"
        )
        found_mismatch = True
    for anc_lfn in from_ancestors_lfns:
        if anc_lfn not in input_lfns:
            output_info["extra_ancestors"].append(anc_lfn)
            found_mismatch = True
    for in_lfn in input_lfns:
        if in_lfn not in from_ancestors_lfns:
            output_info["missing_ancestors"].append(in_lfn)
            found_mismatch = True
    # Can check here too that files have not been run-over multiple times
    dupes = []
    if len(set(from_ancestors_lfns)) != len(from_ancestors_lfns):
        # now we know we have duplicates find out which ones
        no_dupes = set()
        prev_len = 0
        for lfn in from_ancestors_lfns:
            no_dupes.add(lfn)
            if len(no_dupes) == prev_len:
                dupes.append(lfn)
            else:
                prev_len += 1
        gLogger.warn("Found LFN(s) that have been run-over more than once!")
    output_info["duplicate_ancestors"] = dupes
    if not found_mismatch:
        gLogger.always("All checks passed!")
    return output_info


def _get_total_num_events(lfns, bk):
    """
    `lfns` can be a list of dicts with `lfn` and `n_events` as keys or
    just a list of LFN strings and it will lookup the number of events
    """
    total_events = 0
    no_events_lfns = []
    lfns_to_sum = []
    # if a dict of lfns we may already have EventStat/n_events
    bad_input = False
    if all([isinstance(entry, dict) for entry in lfns]):
        if all(["n_events" in entry for entry in lfns]):
            lfns_to_sum = lfns
        else:
            bad_input = True
    elif all([isinstance(entry, str) for entry in lfns]):
        lfns_meta_success = {}
        for lfn_batch in track(
            _batched(lfns, 10_000), total=len(lfns) // 10_000, description="Getting file metadata from input query"
        ):
            lfns_meta = returnValueOrRaise(bk.getFileMetadata(list(lfn_batch)))
            if lfns_meta["Failed"] != []:
                raise RuntimeError(f"Failed to get all metadata for LFNs: {lfns_meta['Failed']}")
            lfns_meta_success |= lfns_meta["Successful"]
        lfns_to_sum = [{"lfn": lfn, "n_events": lfn_info["EventStat"]} for lfn, lfn_info in lfns_meta_success.items()]
    else:
        bad_input = True
    if bad_input:
        raise NotImplementedError(
            "Must provide a list of dictionaries with the fields `lfn` and\
            `n_events` or just a list of LFN names"
        )

    for lfn_info in lfns_to_sum:
        events = lfn_info["n_events"]
        # Seems there are some files with EventStat = 0 but still have a
        # Luminosity value in data
        if events == 0 or events is None:
            no_events_lfns.append(lfn_info["lfn"])
        else:
            total_events += events
    return total_events, no_events_lfns


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import os

import DIRAC
from DIRAC.tests.Utilities.utils import find_all

from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob


def createJob(local=True, workspace=os.path.dirname(DIRAC.__file__)):
    gaudirunJob = LHCbJob()
    gaudirunJob.setName("gaudirun-Gauss-test-wrong-config-will-fail")

    workspace = os.environ.get("WORKSPACE", workspace)
    inputSandbox = [
        find_all("prodConf_Gauss_00012345_00067890_1.py", workspace, "/tests/System/GridTestSubmission")[0],
        find_all("wrongConfig.cfg", workspace, "/tests/System/GridTestSubmission")[0],
    ]
    if not local:
        pilot_cfg = find_all("pilot.cfg", workspace + "/PilotInstallDIR")
        if not pilot_cfg:
            pilot_cfg = find_all("pilot.cfg", workspace)[0]
        inputSandbox += [pilot_cfg]
    gaudirunJob.setInputSandbox(inputSandbox)

    gaudirunJob.setOutputSandbox("00012345_00067890_1.sim")

    optGauss = "$APPCONFIGOPTS/Gauss/Sim08-Beam3500GeV-md100-2011-nu2.py;"
    optDec = "$DECFILESROOT/options/34112104.py;"
    optPythia = "$LBPYTHIAROOT/options/Pythia.py;"
    optOpts = "$APPCONFIGOPTS/Gauss/G4PL_FTFP_BERT_EmNoCuts.py;"
    optCompr = "$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;"
    optPConf = "prodConf_Gauss_00012345_00067890_1.py"
    options = optGauss + optDec + optPythia + optOpts + optCompr + optPConf

    gaudirunJob.setApplication(
        "Gauss",
        "v45r5",
        options,
        extraPackages="AppConfig.v3r171;Gen/DecFiles.v27r14p1;ProdConf.v1r9",
        systemConfig="x86_64-slc5-gcc43-opt",
        modulesNameList=[
            "CreateDataFile",
            "GaudiApplication",
            "FileUsage",
            "UploadOutputData",
            "UploadLogFile",
            "FailoverRequest",
            "UserJobFinalization",
        ],
        parametersList=[
            ("applicationName", "string", "", "Application Name"),
            ("applicationVersion", "string", "", "Application Version"),
            ("applicationLog", "string", "", "Name of the output file of the application"),
            ("optionsFile", "string", "", "Options File"),
            ("extraOptionsLine", "string", "", "This is appended to standard options"),
            ("inputDataType", "string", "", "Input Data Type"),
            ("inputData", "string", "", "Input Data"),
            ("numberOfEvents", "string", "", "Events treated"),
            ("multiCore", "string", "", "If the step can run multicore"),
            ("StopSigNumber", "string", "", "signal number (interpreted by Gaudi)"),
            ("StopSigRegex", "string", "", "RegEx of what to stop"),
            ("extraPackages", "string", "", "ExtraPackages"),
            ("listoutput", "list", [], "StepOutputList"),
            ("SystemConfig", "string", "", "binary tag"),
            ("eventTimeout", "string", "", "Event timeout"),
        ],
        eventTimeout=1000,
    )

    gaudirunJob._addParameter(gaudirunJob.workflow, "PRODUCTION_ID", "string", "00012345", "ProductionID")
    gaudirunJob._addParameter(gaudirunJob.workflow, "JOB_ID", "string", "00067890", "JobID")
    gaudirunJob._addParameter(gaudirunJob.workflow, "configName", "string", "testCfg", "ConfigName")
    gaudirunJob._addParameter(gaudirunJob.workflow, "configVersion", "string", "testVer", "ConfigVersion")
    outputList = [
        {
            "stepName": "GaussStep1",
            "outputDataType": "sim",
            "outputBKType": "SIM",
            "outputDataSE": "Tier1_MC-DST",
            "outputDataName": "00012345_00067890_1.sim",
        }
    ]
    gaudirunJob._addParameter(gaudirunJob.workflow, "outputList", "list", outputList, "outputList")
    gaudirunJob._addParameter(gaudirunJob.workflow, "outputDataFileMask", "string", "", "outputFM")
    gaudirunJob._addParameter(gaudirunJob.workflow, "outputMode", "string", "Local", "OM")
    gaudirunJob._addParameter(gaudirunJob.workflow, "LogLevel", "string", "DEBUG", "LL")
    outputFilesDict = [
        {"outputDataName": "00012345_00067890_1.sim", "outputDataSE": "Tier1_MC-DST", "outputDataType": "SIM"}
    ]
    gaudirunJob._addParameter(
        gaudirunJob.workflow.step_instances[0], "listoutput", "list", outputFilesDict, "listoutput"
    )

    gaudirunJob.setLogLevel("DEBUG")
    gaudirunJob.setDIRACPlatform()
    if local:
        gaudirunJob.setConfigArgs("pilot.cfg wrongConfig.cfg")
    else:
        gaudirunJob.setConfigArgs("wrongConfig.cfg")

    gaudirunJob.setCPUTime(172800)

    return gaudirunJob



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_TS_Client_TaskManager."""
import unittest

import LHCbDIRAC.TransformationSystem.Client.TaskManager as moduleTested

################################################################################


class TaskManager_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""
        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.LHCbWorkflowTasks

    def tearDown(self):
        """TearDown."""
        del self.testClass
        del self.moduleTested


################################################################################
# Tests


class TaskManager_Success(TaskManager_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""
        instance = self.testClass()
        self.assertEqual("LHCbWorkflowTasks", instance.__class__.__name__)


################################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for:
LHCbDIRAC.HCbDIRAC.Core.Utilities.XMLSummaries."""
import pytest
from LHCbDIRAC.Core.Utilities.XMLSummaries import xmltojsonCat1, xmltojsonCat2, xmltojsonCat3, difisnotnull, ranges


lCategory1 = [
    {"@name": "MCVeloHitPacker/# PackedData", "#text": "244708"},
    {"@name": "MCPuVetoHitPacker/# PackedData", "#text": "9800"},
    {"@name": "MCTTHitPacker/# PackedData", "#text": "73624"},
    {"@name": "MCITHitPacker/# PackedData", "#text": "86436"},
]
dictCategory1 = {
    "MCITHitPacker": {"PackedData": 86436},
    "MCPuVetoHitPacker": {"PackedData": 9800},
    "MCTTHitPacker": {"PackedData": 73624},
    "MCVeloHitPacker": {"PackedData": 244708},
}


@pytest.mark.parametrize("test_input,expected", [(lCategory1, dictCategory1), ([], {})])
def test_xmltojsonCat1(test_input, expected):
    assert xmltojsonCat1(test_input) == expected


lCategory2 = [
    {"@name": "TTHitMonitor/DeltaRay", "#text": "10184"},
    {"@name": "TTHitMonitor/betaGamma", "#text": "72641099"},
    {"@name": "TTHitMonitor/numberHits", "#text": "73624"},
    {"@name": "ITHitMonitor/DeltaRay", "#text": "4208"},
    {"@name": "ITHitMonitor/betaGamma", "#text": "224730238"},
    {"@name": "ITHitMonitor/numberHits", "#text": "86436"},
    {"@name": "OTHitMonitor/DeltaRay", "#text": "16893"},
    {"@name": "OTHitMonitor/betaGamma", "#text": "338754733"},
    {"@name": "OTHitMonitor/numberHits", "#text": "253748"},
]

dictCategory2 = {
    "ITHitMonitor": {"DeltaRay": 4208, "betaGamma": 224730238, "numberHits": 86436},
    "OTHitMonitor": {"DeltaRay": 16893, "betaGamma": 338754733, "numberHits": 253748},
    "TTHitMonitor": {"DeltaRay": 10184, "betaGamma": 72641099, "numberHits": 73624},
}


@pytest.mark.parametrize("test_input,expected", [(lCategory2, dictCategory2), ([], {})])
def test_xmltojsonCat2(test_input, expected):
    assert xmltojsonCat2(test_input) == expected


lCategory3 = [
    {"@name": "CheckPuVetoHits/Diff.    - Displacement x", "#text": "0"},
    {"@name": "CheckPuVetoHits/Diff.    - Displacement y", "#text": "0"},
    {"@name": "CheckPuVetoHits/Diff.    - Displacement z", "#text": "0"},
    {"@name": "CheckPuVetoHits/Diff.    - Energy", "#text": "0"},
    {"@name": "CheckPuVetoHits/Diff.    - Entry Point x", "#text": "0"},
    {"@name": "CheckPuVetoHits/Diff.    - Entry Point y", "#text": "0"},
    {"@name": "CheckPuVetoHits/Diff.    - Entry Point z", "#text": "0"},
    {"@name": "CheckPuVetoHits/Diff.    - Parent |P|", "#text": "3"},
    {"@name": "CheckPuVetoHits/Diff.    - TOF", "#text": "0"},
]
dictCategory3 = {
    "CheckPuVetoHits/Diff.": {
        "Displacement": {"x": 0, "y": 0, "z": 0},
        "Energy": 0,
        "Entry Point": {"x": 0, "y": 0, "z": 0},
        "Parent |P|": 3,
        "TOF": 0,
    }
}


@pytest.mark.parametrize("test_input,expected", [(lCategory3, dictCategory3), ([], {})])
def test_xmltojsonCat3(test_input, expected):
    assert xmltojsonCat3(test_input) == expected


dictCategory3NotNull = {
    "CheckPuVetoHits/Diff.": {
        "Displacement": {"x": 0, "y": 0, "z": 0},
        "Energy": 0,
        "Entry Point": {"x": 0, "y": 0, "z": 0},
        "Parent |P|": 3,
        "TOF": 0,
    }
}

dictCategory3Null = {
    "CheckRichOpPhot/Diff.": {
        "Cherenkov": {"Phi": 0, "Theta": 0},
        "Emission Point": {"x": 0, "y": 0, "z": 0},
        "Energy": 0,
        "HPD In. Point": {"x": 0, "y": 0, "z": 0},
        "HPD QW Point": {"x": 0, "y": 0, "z": 0},
        "Parent Momentum": {"x": 0, "y": 0, "z": 0},
        "Prim. Mirr.": {"x": 0, "y": 0, "z": 0},
        "Sec. Mirr.": {"x": 0, "y": 0, "z": 0},
    }
}


@pytest.mark.parametrize("test_input,expected", [(dictCategory3NotNull, True), (dictCategory3Null, False)])
def test_difisnotnull(test_input, expected):
    assert difisnotnull(test_input) == expected


def test_ranges():
    testList = [0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 12, 13, 14, 20, 21, 22, 23]

    assert ranges(testList) == [0, 8, 10, 14, 20, 23]
